"""
init
"""

default_app_config: str = "fleetfinder.apps.FleetFinderConfig"

github_url = "https://github.com/ppfeufer/aa-fleetfinder"

__title__ = "Fleet Finder"
__version__ = "0.1.0-alpha.9"

__verbose_name__ = "Fleet Finder for Alliance Auth"
__user_agent_name__ = "Fleet-Finder-for-Alliance-Auth"

__user_agent__ = f"{__verbose_name__} v{__version__} {github_url}"



"""
tasks
"""

# Standard Library
from concurrent.futures import ThreadPoolExecutor, as_completed

# Third Party
from bravado.exception import (
    HTTPBadGateway,
    HTTPGatewayTimeout,
    HTTPNotFound,
    HTTPServiceUnavailable,
)
from celery import shared_task

# Django
from django.core.cache import cache
from django.utils import timezone

# Alliance Auth
from allianceauth.eveonline.models import EveCharacter
from allianceauth.services.hooks import get_extension_logger
from allianceauth.services.tasks import QueueOnce
from esi.models import Token

# AA Fleet Finder
from fleetfinder import __title__
from fleetfinder.models import Fleet
from fleetfinder.providers import esi
from fleetfinder.utils import LoggerAddTag

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


ESI_ERROR_LIMIT = 50
ESI_TIMEOUT_ONCE_ERROR_LIMIT_REACHED = 60
ESI_MAX_RETRIES = 3


CACHE_KEY_NOT_IN_FLEET_ERROR = (
    "fleetfinder_task_check_fleet_adverts_error_counter_not_in_fleet_"
)
CACHE_KEY_NO_FLEET_ERROR = (
    "fleetfinder_task_check_fleet_adverts_error_counter_no_fleet_"
)
CACHE_KEY_FLEET_CHANGED_ERROR = (
    "fleetfinder_task_check_fleet_adverts_error_counter_fleet_changed_"
)
CACHE_MAX_ERROR_COUNT = 3


# params for all tasks
TASK_DEFAULT_KWARGS = {
    "time_limit": 120,  # stop after 2 minutes
}

# params for tasks that make ESI calls
TASK_ESI_KWARGS = {
    **TASK_DEFAULT_KWARGS,
    **{
        "autoretry_for": (
            OSError,
            HTTPBadGateway,
            HTTPGatewayTimeout,
            HTTPServiceUnavailable,
        ),
        "retry_kwargs": {"max_retries": ESI_MAX_RETRIES},
        "retry_backoff": True,
    },
}


@shared_task
def open_fleet(character_id, motd, free_move, name, groups):
    """
    open a fleet
    :param character_id:
    :param motd:
    :param free_move:
    :param name:
    :param groups:
    :return:
    """

    required_scopes = ["esi-fleets.read_fleet.v1", "esi-fleets.write_fleet.v1"]

    esi_client = esi.client
    token = Token.get_token(character_id, required_scopes)

    fleet_result = esi_client.Fleets.get_characters_character_id_fleet(
        character_id=token.character_id, token=token.valid_access_token()
    ).result()
    fleet_id = fleet_result.pop("fleet_id")
    fleet_role = fleet_result.pop("role")

    if fleet_id is None or fleet_role is None or fleet_role != "fleet_commander":
        return

    fleet_commander = EveCharacter.objects.get(character_id=token.character_id)

    fleet = Fleet(
        fleet_id=fleet_id,
        created_at=timezone.now(),
        motd=motd,
        is_free_move=free_move,
        fleet_commander=fleet_commander,
        name=name,
    )
    fleet.save()
    fleet.groups.set(groups)

    esi_fleet = {"is_free_move": free_move, "motd": motd}
    esi_client.Fleets.put_fleets_fleet_id(
        fleet_id=fleet_id, token=token.valid_access_token(), new_settings=esi_fleet
    ).result()


@shared_task
def send_fleet_invitation(character_ids, fleet_id):
    """
    send a fleet invitation through the eve client
    :param character_ids:
    :param fleet_id:
    """

    required_scopes = ["esi-fleets.write_fleet.v1"]
    fleet = Fleet.objects.get(fleet_id=fleet_id)
    fleet_commander_token = Token.get_token(
        fleet.fleet_commander.character_id, required_scopes
    )
    _processes = []

    with ThreadPoolExecutor(max_workers=50) as ex:
        for _chracter_id in character_ids:
            _processes.append(
                ex.submit(
                    send_invitation,
                    character_id=_chracter_id,
                    fleet_commander_token=fleet_commander_token,
                    fleet_id=fleet_id,
                )
            )

    for item in as_completed(_processes):
        _ = item.result()


@shared_task
def send_invitation(character_id, fleet_commander_token, fleet_id):
    """
    open the fleet invite window in the eve client
    :param character_id:
    :param fleet_commander_token:
    :param fleet_id:
    """

    esi_client = esi.client
    invitation = {"character_id": character_id, "role": "squad_member"}

    esi_client.Fleets.post_fleets_fleet_id_members(
        fleet_id=fleet_id,
        token=fleet_commander_token.valid_access_token(),
        invitation=invitation,
    ).result()


def close_esi_fleet(fleet: Fleet, reason: str) -> None:
    """
    closing registered fleet
    :param fleet:
    :param reason:
    """

    logger.info(
        "Closing fleet with ID {fleet_id}. Reason: {reason}".format(
            fleet_id=fleet.fleet_id, reason=reason
        )
    )

    fleet.delete()


def esi_fleetadvert_error_handling(
    cache_key: str, fleet: Fleet, logger_message: str
) -> None:
    """
    ESI error handling
    :param cache_key:
    :param fleet:
    :param logger_message:
    """

    if int(cache.get(cache_key + str(fleet.fleet_id))) < CACHE_MAX_ERROR_COUNT:
        error_count = int(cache.get(cache_key + str(fleet.fleet_id)))

        error_count += 1

        logger.info(
            '"{logger_message}" Error Count: {error_count}.'.format(
                logger_message=logger_message, error_count=error_count
            )
        )

        cache.set(
            cache_key + str(fleet.fleet_id),
            str(error_count),
            75,
        )
    else:
        close_esi_fleet(fleet=fleet, reason=logger_message)


def init_error_caches(fleet: Fleet) -> None:
    """
    initialize error caches
    :param fleet:
    """

    if cache.get(CACHE_KEY_FLEET_CHANGED_ERROR + str(fleet.fleet_id)) is None:
        cache.set(CACHE_KEY_FLEET_CHANGED_ERROR + str(fleet.fleet_id), "0", 75)

    if cache.get(CACHE_KEY_NO_FLEET_ERROR + str(fleet.fleet_id)) is None:
        cache.set(CACHE_KEY_NO_FLEET_ERROR + str(fleet.fleet_id), "0", 75)

    if cache.get(CACHE_KEY_NOT_IN_FLEET_ERROR + str(fleet.fleet_id)) is None:
        cache.set(CACHE_KEY_NOT_IN_FLEET_ERROR + str(fleet.fleet_id), "0", 75)


@shared_task(**{**TASK_ESI_KWARGS}, **{"base": QueueOnce})
def check_fleet_adverts():
    """
    scheduled task
    check for fleets adverts
    """

    required_scopes = ["esi-fleets.read_fleet.v1", "esi-fleets.write_fleet.v1"]

    esi_client = esi.client

    fleets = Fleet.objects.all()

    fleet_count = fleets.count()

    logger.info(
        "{fleet_count} registered fleets found. {processing_text}".format(
            fleet_count=fleet_count,
            processing_text="Processing ..."
            if fleet_count > 0
            else "Nothing to do ...",
        )
    )

    if fleet_count > 0:
        for fleet in fleets:
            init_error_caches(fleet=fleet)

            logger.info(
                "Processing information for fleet with ID {fleet_id}".format(
                    fleet_id=fleet.fleet_id
                )
            )

            token = Token.get_token(fleet.fleet_commander.character_id, required_scopes)

            try:
                fleet_result = esi_client.Fleets.get_characters_character_id_fleet(
                    character_id=token.character_id, token=token.valid_access_token()
                ).result()

                fleet_id = fleet_result["fleet_id"]

                if fleet_id != fleet.fleet_id:
                    esi_fleetadvert_error_handling(
                        cache_key=CACHE_KEY_FLEET_CHANGED_ERROR,
                        fleet=fleet,
                        logger_message="FC switched to another fleet",
                    )

            except HTTPNotFound:
                esi_fleetadvert_error_handling(
                    cache_key=CACHE_KEY_NOT_IN_FLEET_ERROR,
                    fleet=fleet,
                    logger_message=(
                        "FC is not in the registered fleet anymore or "
                        "fleet is no longer available."
                    ),
                )

            except Exception:
                esi_fleetadvert_error_handling(
                    cache_key=CACHE_KEY_NO_FLEET_ERROR,
                    fleet=fleet,
                    logger_message="Registered fleet is no longer available.",
                )


@shared_task
def get_fleet_composition(fleet_id):
    """
    getting the fleet composition
    :param fleet_id:
    :return:
    """

    required_scopes = ["esi-fleets.read_fleet.v1", "esi-fleets.write_fleet.v1"]

    esi_client = esi.client

    fleet = Fleet.objects.get(fleet_id=fleet_id)
    token = Token.get_token(fleet.fleet_commander.character_id, required_scopes)
    fleet_infos = esi_client.Fleets.get_fleets_fleet_id_members(
        fleet_id=fleet_id, token=token.valid_access_token()
    ).result()

    characters = {}
    systems = {}
    ship_type = {}

    for member in fleet_infos:
        characters[member["character_id"]] = ""
        systems[member["solar_system_id"]] = ""
        ship_type[member["ship_type_id"]] = ""

    ids = []
    ids.extend(list(characters.keys()))
    ids.extend(list(systems.keys()))
    ids.extend(list(ship_type.keys()))

    ids_to_name = esi_client.Universe.post_universe_names(ids=ids).result()

    for member in fleet_infos:
        index_character = [x["id"] for x in ids_to_name].index(member["character_id"])
        member["character_name"] = ids_to_name[index_character]["name"]

        index_solar_system = [x["id"] for x in ids_to_name].index(
            member["solar_system_id"]
        )
        member["solar_system_name"] = ids_to_name[index_solar_system]["name"]

        index_ship_type = [x["id"] for x in ids_to_name].index(member["ship_type_id"])
        member["ship_type_name"] = ids_to_name[index_ship_type]["name"]

    aggregate = get_fleet_aggregate(fleet_infos)

    # differential = dict()
    #
    # for key, value in aggregate.items():
    #     fleet_info_agg = FleetInformation.objects.filter(
    #         fleet__fleet_id=fleet_id, ship_type_name=key
    #     )
    #
    #     if fleet_info_agg.count() > 0:
    #         differential[key] = value - fleet_info_agg.latest("date").count
    #     else:
    #         differential[key] = value
    #
    #     FleetInformation.objects.create(fleet=fleet, ship_type_name=key, count=value)

    return FleetViewAggregate(
        fleet_infos,
        aggregate,
        # differential,
    )


@shared_task
def get_fleet_aggregate(fleet_infos):
    """
    getting aggregate numbers for fleet composition
    :param fleet_infos:
    :return:
    """

    counts = dict()

    for member in fleet_infos:
        type_ = member.get("ship_type_name")

        if type_ in counts:
            counts[type_] += 1
        else:
            counts[type_] = 1

    return counts


class FleetViewAggregate:
    """
    helper class
    """

    def __init__(
        self,
        fleet,
        aggregate,
        # differential,
    ):
        self.fleet = fleet
        self.aggregate = aggregate
        # self.differential = differential



# Generated by Django 3.2.9 on 2021-11-15 12:31

# Django
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("fleetfinder", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="fleet",
            name="motd",
            field=models.TextField(blank=True, default=""),
        ),
    ]



"""
URL config
"""

# Django
from django.urls import path

# AA Fleet Finder
from fleetfinder import views

app_name: str = "fleetfinder"

urlpatterns = [
    path("", views.dashboard, name="dashboard"),
    path("create/", views.create_fleet, name="create_fleet"),
    path("save/", views.save_fleet, name="save_fleet"),
    path("join/<int:fleet_id>/", views.join_fleet, name="join_fleet"),
    path("details/<int:fleet_id>/", views.fleet_details, name="fleet_details"),
    path("edit/<int:fleet_id>/", views.edit_fleet, name="edit_fleet"),
    # ajax calls
    path("call/dashboard/", views.ajax_dashboard, name="ajax_dashboard"),
    path(
        "call/details/<int:fleet_id>/",
        views.ajax_fleet_details,
        name="ajax_fleet_details",
    ),
]



"""
views
"""

# Third Party
from bravado.exception import HTTPNotFound

# Django
from django.contrib.auth.decorators import login_required, permission_required
from django.db.models import Q
from django.http import JsonResponse
from django.shortcuts import redirect, render
from django.template.defaulttags import register
from django.urls import reverse
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth.eveonline.evelinks.eveimageserver import character_portrait_url
from allianceauth.eveonline.models import EveCharacter
from allianceauth.groupmanagement.models import AuthGroup
from allianceauth.services.hooks import get_extension_logger
from esi.decorators import token_required

# AA Fleet Finder
from fleetfinder import __title__
from fleetfinder.app_settings import avoid_cdn
from fleetfinder.models import Fleet
from fleetfinder.tasks import get_fleet_composition, open_fleet, send_fleet_invitation
from fleetfinder.utils import LoggerAddTag

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


@login_required()
@permission_required("fleetfinder.access_fleetfinder")
def dashboard(request):
    """
    dashboard view
    :param request:
    :return:
    """

    # groups = request.user.groups.all()
    # fleets = Fleet.objects.filter(Q(groups__group__in=groups) | Q(groups=None)).all()

    context = {
        # "fleets": fleets,
        "avoid_cdn": avoid_cdn(),  # AVOID_CDN setting
    }

    if "error_edit_fleet" in request.session:
        context["error"] = request.session["error_edit_fleet"].get("error", "")

        del request.session["error_edit_fleet"]

    logger.info(f"Module called by {request.user}")

    return render(request, "fleetfinder/dashboard.html", context)


@login_required()
@permission_required("fleetfinder.access_fleetfinder")
def ajax_dashboard(request) -> JsonResponse:
    """
    dashboard view
    :param request:
    :return:
    """

    data = list()
    groups = request.user.groups.all()
    fleets = Fleet.objects.filter(Q(groups__group__in=groups) | Q(groups=None)).all()

    for fleet in fleets:
        fleet_commander_name = fleet.fleet_commander.character_name
        fleet_commander_portrait = (
            '<img class="img-rounded eve-character-portrait" '
            'src="{portrait_url}" '
            'alt="{character_name}">'.format(
                portrait_url=character_portrait_url(
                    character_id=fleet.fleet_commander.character_id, size=32
                ),
                character_name=fleet_commander_name,
            )
        )
        fleet_commander_html = fleet_commander_portrait + fleet_commander_name

        button_join_url = reverse("fleetfinder:join_fleet", args=[fleet.fleet_id])
        button_join = (
            '<a href="{button_url}" '
            'class="btn btn-sm btn-default">{button_text}</a>'.format(
                button_url=button_join_url, button_text=_("Join Fleet")
            )
        )

        button_details = ""
        button_edit = ""

        if request.user.has_perm("fleetfinder.manage_fleets"):
            button_details_url = reverse(
                "fleetfinder:fleet_details", args=[fleet.fleet_id]
            )
            button_details = (
                '<a href="{button_url}" '
                'class="btn btn-sm btn-default">{button_text}</a>'.format(
                    button_url=button_details_url, button_text=_("View fleet details")
                )
            )

            button_edit_url = reverse("fleetfinder:edit_fleet", args=[fleet.fleet_id])
            button_edit = (
                '<a href="{button_url}" '
                'class="btn btn-sm btn-default">{button_text}</a>'.format(
                    button_url=button_edit_url, button_text=_("Edit Fleet advert")
                )
            )

        data.append(
            {
                "fleet_commander": {
                    "html": fleet_commander_html,
                    "sort": fleet_commander_name,
                },
                "fleet_name": fleet.name,
                "created_at": fleet.created_at,
                "join": button_join,
                "details": button_details,
                "edit": button_edit,
            }
        )

    return JsonResponse(data, safe=False)


@login_required()
@permission_required("fleetfinder.manage_fleets")
@token_required(
    scopes=(
        "esi-fleets.read_fleet.v1",
        "esi-fleets.write_fleet.v1",
    )
)
def create_fleet(request, token):
    """
    create fleet view
    :param request:
    :param token:
    :return:
    """

    if request.method == "POST":
        auth_groups = AuthGroup.objects.filter(internal=False).all()

        context = {}

        if "modified_fleet_data" in request.session:
            context["error"] = request.session["modified_fleet_data"].get("error", "")
            context["motd"] = request.session["modified_fleet_data"].get("motd", "")
            context["name"] = request.session["modified_fleet_data"].get("name", "")
            context["groups"] = request.session["modified_fleet_data"].get("groups", "")
            context["is_free_move"] = request.session["modified_fleet_data"].get(
                "free_move", ""
            )
            context["character_id"] = token.character_id
            context["auth_groups"] = auth_groups

            del request.session["modified_fleet_data"]
        else:
            context = {"character_id": token.character_id, "auth_groups": auth_groups}

        # AVOID_CDN setting
        context["avoid_cdn"] = avoid_cdn()

        logger.info(f"Fleet created by {request.user}")

        return render(request, "fleetfinder/create_fleet.html", context=context)

    return redirect("fleetfinder:dashboard")


@login_required()
@permission_required("fleetfinder.manage_fleets")
def edit_fleet(request, fleet_id):
    """
    fleet edit view
    :param request:
    :param fleet_id:
    :return:
    """

    fleet = Fleet.objects.get(fleet_id=fleet_id)
    auth_groups = AuthGroup.objects.filter(internal=False)

    context = {
        "character_id": fleet.fleet_commander.character_id,
        "auth_groups": auth_groups,
        "fleet": fleet,
        "avoid_cdn": avoid_cdn(),  # AVOID_CDN setting
    }

    logger.info(
        "Fleet {fleet_id} edit view by {user}".format(
            fleet_id=fleet_id, user=request.user
        )
    )

    return render(request, "fleetfinder/edit_fleet.html", context=context)


@login_required()
@permission_required("fleetfinder.access_fleetfinder")
def join_fleet(request, fleet_id):
    """
    join fleet view
    :param request:
    :param fleet_id:
    :return:
    """

    context = {}
    groups = request.user.groups.all()
    fleet = Fleet.objects.filter(
        Q(groups__group__in=groups) | Q(groups=None), fleet_id=fleet_id
    ).count()

    if fleet == 0:
        return redirect("fleetfinder:dashboard")

    if request.method == "POST":
        character_ids = request.POST.getlist("character_ids", [])
        send_fleet_invitation.delay(character_ids, fleet_id)

        return redirect("fleetfinder:dashboard")

    characters = (
        EveCharacter.objects.filter(character_ownership__user=request.user)
        .select_related()
        .order_by("character_name")
    )

    context["characters"] = characters

    # AVOID_CDN setting
    context["avoid_cdn"] = avoid_cdn()

    return render(request, "fleetfinder/join_fleet.html", context=context)


@login_required()
@permission_required("fleetfinder.manage_fleets")
def save_fleet(request):
    """
    save fleet
    :param request:
    :return:
    """

    if request.method == "POST":
        free_move = request.POST.get("free_move", False)

        if free_move == "on":
            free_move = True

        motd = request.POST.get("motd", "")
        name = request.POST.get("name", "")
        groups = request.POST.getlist("groups", [])

        try:
            open_fleet(request.POST["character_id"], motd, free_move, name, groups)
        except HTTPNotFound as ex:
            if request.POST.get("origin", "") == "edit":
                # Here ccp return "character not in fleet".
                # Instead push our own message to be clearer
                request.session["error_edit_fleet"] = {
                    "error": "Fleet advert is no longer valid"
                }

                return redirect("fleetfinder:dashboard")

            if request.POST.get("origin", "") == "create":
                request.session["modified_fleet_data"] = {
                    "error": ex.swagger_result["error"],
                    "motd": motd,
                    "name": name,
                    "free_move": free_move,
                    "groups": groups,
                }

                return redirect("fleetfinder:create_fleet")

    return redirect("fleetfinder:dashboard")


@login_required()
@permission_required("fleetfinder.manage_fleets")
def fleet_details(request, fleet_id):
    """
    fleet details view
    :param request:
    :param fleet_id:
    :return:
    """

    # fleet = get_fleet_composition(fleet_id)

    context = {
        # "fleet": fleet,
        # "fleet_details": fleet.fleet,
        # "aggregate": fleet.aggregate,
        # "differential": fleet.differential,
        "avoid_cdn": avoid_cdn(),  # AVOID_CDN setting
        "fleet_id": fleet_id,
    }

    logger.info(
        "Fleet {fleet_id} details view called by {user}".format(
            fleet_id=fleet_id, user=request.user
        )
    )

    return render(request, "fleetfinder/fleet_details.html", context=context)


@login_required()
@permission_required("fleetfinder.manage_fleets")
def ajax_fleet_details(request, fleet_id) -> JsonResponse:
    """

    :param request:
    :param fleet_id:
    """

    fleet = get_fleet_composition(fleet_id)

    data = {
        "fleet_member": [],
        # "differential": fleet.differential,
        "fleet_composition": [],
    }

    # data = {"differential": [fleet.differential]}
    for member in fleet.fleet:
        data["fleet_member"].append(member)

    for ship, number in fleet.aggregate.items():
        data["fleet_composition"].append({"ship_type_name": ship, "number": number})

    return JsonResponse(data, safe=False)


@register.filter
def get_item(dictionary, key):
    """
    little helper: get a key from a dictionary
    :param dictionary:
    :param key:
    :return:
    """

    return dictionary.get(key)
